import pathlib
from bootplot import bootplot
from bootplot.base import merge_images
import numpy as np
import shutil


def make_dataset():
    return np.array([11, 7, 7, 5, 6, 5, 4, 4, 7, 9])


def make_bar_chart(data_subset, data_full, ax):
    n_classes = 15
    bar_ticks, bar_counts = np.unique(data_subset, return_counts=True)
    bar_ticks_adjusted = np.arange(n_classes)
    bar_counts_adjusted = np.zeros_like(bar_ticks_adjusted, dtype=np.float32)
    bar_counts_adjusted[bar_ticks] = bar_counts
    ax.bar(bar_ticks_adjusted, bar_counts_adjusted)


def make_directory():
    directory = pathlib.Path('./tests_tmp').absolute()
    if directory.exists():
        shutil.rmtree(directory)
    directory.mkdir()
    return directory


def test_png():
    # should generate png
    directory = make_directory()
    image_path = directory / 'test.png'
    x = make_dataset()
    bootplot(make_bar_chart, x, m=5, output_image_path=image_path)
    assert image_path.exists()
    shutil.rmtree(directory)


def test_jpg():
    # should generate jpg
    directory = make_directory()
    image_path = directory / 'test.jpg'
    x = make_dataset()
    bootplot(make_bar_chart, x, m=5, output_image_path=image_path)
    assert image_path.exists()
    shutil.rmtree(directory)


def test_gif():
    # should generate gif
    directory = make_directory()
    animation_path = directory / 'test.gif'
    x = make_dataset()
    bootplot(make_bar_chart, x, m=5, output_animation_path=animation_path)
    assert animation_path.exists()
    shutil.rmtree(directory)


def test_mp4():
    # should generate mp4
    directory = make_directory()
    animation_path = directory / 'test.mp4'
    x = make_dataset()
    bootplot(make_bar_chart, x, m=5, output_animation_path=animation_path)
    assert animation_path.exists()
    shutil.rmtree(directory)


def test_bmp():
    # should generate bmp
    directory = make_directory()
    image_path = directory / 'test.bmp'
    x = make_dataset()
    bootplot(make_bar_chart, x, m=5, output_image_path=image_path)
    assert image_path.exists()
    shutil.rmtree(directory)


def test_merge():
    np.random.seed(0)
    images = np.random.randint(low=0, high=256, size=(25, 100, 100, 4))
    merged = merge_images(images)
    assert merged.shape == (100, 100, 4)
    assert np.min(merged) >= 0
    assert np.max(merged) <= 255
    assert merged.dtype == np.uint8



from bootplot.sorting import PCASorter, DefaultSorter, HorizontalMassSorter, TravelingSalesmanSorter
import numpy as np


def make_gray_images():
    np.random.seed(0)
    return np.random.uniform(low=0, high=1, size=(25, 100, 100))


def test_pca():
    images = make_gray_images()
    sorter = PCASorter()

    order = sorter.sort(images, features="selected")
    assert len(order) == len(images)
    assert len(set(order)) == len(images)
    assert set(order) == set(range(len(images)))

    order = sorter.sort(images, features="center_mass")
    assert len(order) == len(images)
    assert len(set(order)) == len(images)
    assert set(order) == set(range(len(images)))

    order = sorter.sort(images, features="full")
    assert len(order) == len(images)
    assert len(set(order)) == len(images)
    assert set(order) == set(range(len(images)))


def test_default():
    images = make_gray_images()
    sorter = DefaultSorter()
    order = sorter.sort(images)
    assert order == list(range(len(images)))


def test_hm():
    images = make_gray_images()
    sorter = HorizontalMassSorter()

    order = sorter.sort(images)
    assert len(order) == len(images)
    assert len(set(order)) == len(images)
    assert set(order) == set(range(len(images)))


def test_tsp():
    images = make_gray_images()
    sorter = TravelingSalesmanSorter()

    order = sorter.sort(images)
    assert len(order) == len(images)
    assert len(set(order)) == len(images)
    assert set(order) == set(range(len(images)))



import numpy as np
import scipy.ndimage
from skimage.feature import hog
import cv2


def compute_hog_features(gray_images: np.ndarray) -> np.ndarray:
    """
    Compute features based on histograms of oriented gradients (HOG) for a given set of images.

    :param gray_images: input images.
    :return: features for each image.
    """
    hog_features = np.stack([hog(im, cells_per_block=(1, 1), pixels_per_cell=(10, 10)) for im in gray_images])
    hog_features = (hog_features - np.min(hog_features)) / (np.max(hog_features) - np.min(hog_features))
    return hog_features


def compute_center_of_mass(gray_images: np.ndarray) -> np.ndarray:
    """
    Compute centers of mass for all input images.

    :param gray_images: input images.
    :return: center of mass for each image.
    """
    centers_of_mass = np.zeros((len(gray_images), 2))
    for i in range(len(gray_images)):
        centers_of_mass[i] = scipy.ndimage.center_of_mass(gray_images[i])
    centers_of_mass /= np.array(gray_images.shape[1:])
    return centers_of_mass


def compute_center_of_mass_per_component(gray_images: np.ndarray, threshold: float = 1e-5) -> np.ndarray:
    """
    Compute centers of mass for all input images.
    For each image, we first binarize it, determine its components, then compute the center mass for each component.

    :param gray_images: input images.
    :param threshold: binarization threshold.
    :return: center of mass for each component of each image.
    """
    movement_mask = (np.var(gray_images, axis=0) > threshold).astype(np.uint8)
    n_components, label_mask, stats, centroids = cv2.connectedComponentsWithStats(movement_mask)
    centers = np.zeros((len(gray_images), 2 * n_components - 2))
    for j, image in enumerate(gray_images):
        for i in range(1, n_components):
            center_of_mass = scipy.ndimage.center_of_mass(image, label_mask, i)
            centers[j, i * 2 - 2:i * 2] = center_of_mass
    centers /= np.tile(np.array(gray_images.shape[1:]), n_components - 1)
    return centers



from pathlib import Path
from typing import Union, Tuple

import numpy as np
import imageio
import io
import matplotlib.pyplot as plt
from tqdm import tqdm
from PIL import Image

from bootplot.sorting import sort_images


def fig_to_array(fig: plt.Figure,
                 ax: plt.Axes) -> np.array:
    """
    Retrieve array of pixels from a figure.
    The Axis is cleared afterwards.

    :param fig: figure with the plot. The figure should contain a single Axes object.
    :param ax: axis with the plot.
    :return: numpy array of pixels of the plot.
    """
    with io.BytesIO() as buff:
        fig.savefig(buff, format='raw')
        buff.seek(0)
        data = np.frombuffer(buff.getvalue(), dtype=np.uint8)
    w, h = fig.canvas.get_width_height()
    im = data.reshape((int(h), int(w), -1))
    ax.cla()
    return im


def plot_to_array(plot_function: callable,
                  data: np.ndarray,
                  indices: np.ndarray,
                  fig: plt.Figure,
                  ax: plt.Axes,
                  xlim: Tuple[float, float] = (None, None),
                  ylim: Tuple[float, float] = (None, None),
                  **kwargs) -> np.ndarray:
    """
    Plot data and obtain image.

    :param plot_function: function to do the plotting. The function should receive the data subset, original data, Axes
        object and optional keyword arguments.
    :param data: full data to be used in plotting.
    :param indices: bootstrap resampled indices of the data.
    :param fig: Figure object with a single Axes.
    :param ax: Axes object.
    :param xlim: x axis limits.
    :param ylim: y axis limits.
    :param kwargs: keyword arguments to plot_function.
    :return: image.
    """
    plot_function(data[indices], data, ax, **kwargs)
    ax.set_xlim(*xlim)
    ax.set_ylim(*ylim)
    data = fig_to_array(fig, ax)
    return data


def merge_images(images: np.ndarray) -> np.ndarray:
    """
    Merge images into a static image (averaged image).
    The shape of images is (batch_size, width, height, channels).
    This operation overwrites input images.

    :param images: images corresponding to different bootstrap resamples.
    :return: merged image.
    """
    images = images.astype(np.float32) / 255  # Cast to float
    merged = np.mean(images, axis=0)
    merged = (merged * 255).astype(np.uint8)
    return merged


def decay_images(images: np.ndarray,
                 m: int,
                 decay_length: int) -> np.ndarray:
    """
    Apply visual decay to images.
    Once applied, images[t] will contain a weighted sum of images from t - decay_length to t.

    :param images: array of images corresponding to different bootstrap samples.
    :param m: number of bootstrap samples.
    :param decay_length: consider this many preceding images when creating a decayed image.
    :return: decayed images with the same shape as input images.
    """
    decayed_images = np.zeros((m, *images[0].shape), dtype=np.uint8)
    for i in range(m):
        matrix_indices = np.arange(i - decay_length, i)  # Getting frames at the end makes the gif loop smoothly
        weights = np.arange(1, decay_length + 1)
        weights = weights ** 2
        weights = weights / np.sum(weights)
        weights = weights.reshape(-1, 1, 1, 1)
        decayed_images[i] = (np.sum(images[matrix_indices].astype(np.float32) * weights, axis=0)).astype(np.uint8)
    return decayed_images


def bootplot(f: callable,
             data: np.ndarray,
             m: int = 100,
             output_size_px: Tuple[int, int] = (512, 512),
             output_image_path: Union[str, Path] = None,
             output_animation_path: Union[str, Path] = None,
             sort_type: str = 'tsp',
             sort_kwargs: dict = None,
             decay: int = 0,
             fps: int = 60,
             xlim: Tuple[float, float] = (None, None),
             ylim: Tuple[float, float] = (None, None),
             verbose: bool = False) -> np.ndarray:
    """
    Create a bootstrapped plot.

    :param f: function to perform the plotting. The function should receive the data subset, original data, Axes object.
    :param data: data to be used in plotting.
    :param m: number of boostrap resamples.
    :param output_size_px: output size (height, width) in pixels.
    :param output_image_path: path where the image should be stored. If None, the image is not stored.
    :param output_animation_path: path where the animation should be stored. If None, the animation is not created.
    :param sort_type: method to sort images when constructing the animation. Should be one of the following:
        "tsp" (traveling salesman method on the image similarity graph), "pca" (image projection onto the real line
        using PCA), "hm" (order using center mass in the horizontal direction), "none" (no sorting; random order).
    :param sort_kwargs: keyword arguments for the sorting method. See bootplot.sorting.sort_images for details.
    :param decay: decay length when creating the animation. If 0, no decay is applied.
    :param fps: desired output framerate for the animation.
    :param xlim: x axis limits.
    :param ylim: y axis limits.
    :param verbose: if True, print progress messages.
    :return: bootstrapped plot as a numpy array.
    """
    px_size_inches = 1 / plt.rcParams['figure.dpi']
    fig, ax = plt.subplots(figsize=(output_size_px[0] * px_size_inches, output_size_px[1] * px_size_inches))
    image_samples = np.stack([
        plot_to_array(f, data, np.random.randint(low=0, high=len(data), size=len(data)), fig, ax, xlim=xlim, ylim=ylim)
        for _ in tqdm(range(m), desc='Generating plots', disable=not verbose)
    ])
    merged_image = merge_images(image_samples)[..., :3]  # Do not use the alpha channel
    plt.close(fig)

    if output_image_path is not None:
        if verbose:
            print(f'Saving bootstrapped image to {output_image_path}')
        Image.fromarray(merged_image).save(output_image_path)
    if output_animation_path is not None:
        sort_kwargs = dict() if sort_kwargs is None else sort_kwargs
        order = sort_images(image_samples, sort_type, verbose=verbose, **sort_kwargs)
        order.extend(order[:-1][::-1])  # go in reverse
        order = np.array(order)
        image_samples = image_samples[order]

        # Apply decay
        if decay > 0:
            image_samples = decay_images(image_samples, m=m, decay_length=decay)

        imageio.mimwrite(output_animation_path, image_samples, fps=fps)
        if verbose:
            print(f'Saving bootstrapped animation to {output_animation_path}')

    return merged_image



import abc
from typing import List, Tuple
import scipy.spatial
import scipy.ndimage
import skimage.transform
import cv2
import numpy as np
from sklearn.decomposition import PCA
import scipy
import networkx as nx

from bootplot.image_features import compute_hog_features, compute_center_of_mass, compute_center_of_mass_per_component


class Sorter(abc.ABC):
    def __init__(self, verbose: bool = False):
        """
        Abstract class for sorting.

        :param verbose: if True, print additional information during sorting.
        """
        self.verbose = verbose

    @abc.abstractmethod
    def sort(self, images: np.ndarray, **kwargs) -> List[int]:
        """
        Sort images.

        :param images: input images.
        :param kwargs: keyword arguments for the sorting function.
        :return:
        """
        pass

    def verbose_print(self, *args, **kwargs):
        if self.verbose:
            print(*args, **kwargs)


class DefaultSorter(Sorter):
    def sort(self, images: np.ndarray, **kwargs) -> List[int]:
        """
        Does not sort images.

        :param images: input images.
        :param kwargs: unused.
        :return: indices of images in the original order.
        """
        return list(range(len(images)))  # Does not sort


class PCASorter(Sorter):
    def sort(self, gray_images: np.ndarray, features: str = "center_mass", **kwargs) -> List[int]:
        """
        Sort images with a PCA-based method.
        We first compute image features, then project these features into 1D.
        The order of the projection is the order of the images.

        :param gray_images: input images.
        :param features: type of features to be used for PCA. Must be one of "selected" (HOG, center mass, per component
            center mass), "center_mass" (center mass), "full" (flattened images).
        :param kwargs: unused.
        :return: image indices in for the new ordering.
        """
        if features == "selected":
            hf = compute_hog_features(gray_images)
            cm = compute_center_of_mass(gray_images)
            cmc = compute_center_of_mass_per_component(gray_images)
            features = np.c_[hf, cm, cmc]
        elif features == "center_mass":
            features = compute_center_of_mass(gray_images)
        elif features == "full":
            # Sort images according to 1D PCA projections of their centers of mass
            features = gray_images.reshape(len(gray_images), -1)
        else:
            raise ValueError(f"Unsupported features: '{features}'")
        model = PCA(n_components=1)
        projections = model.fit_transform(features).ravel()
        order = list(np.argsort(projections))
        return order


class TravelingSalesmanSorter(Sorter):
    def distance_matrix(self, gray_images: np.ndarray, features: str = "full") -> np.ndarray:
        """
        Compute distances for all pairs input images. The distances are based on image representations, given by
        the features parameter.

        :param gray_images: input images.
        :param features: type of features to be used for distance computation. Must be one of "selected" (HOG,
            center mass, per component center mass), "center_mass" (center mass), "full" (flattened images).
        :return: square pairwise distance matrix.
        """
        m = len(gray_images)
        if features == 'full':
            ksize = list(gray_images[0].shape[:2])
            ksize[0] += ksize[0] % 2 - 1
            ksize[1] += ksize[1] % 2 - 1
            ksize = ksize[::-1]

            gray_images = np.array([
                cv2.GaussianBlur(im, ksize=tuple(ksize), sigmaX=ksize[0] / 6, sigmaY=ksize[1] / 6)
                for im in gray_images
            ])
            # convolve with a large kernel so that elements are "spread out"
            bootstrapped_vectors = gray_images.reshape(m, -1)
            distance_matrix = scipy.spatial.distance.cdist(bootstrapped_vectors, bootstrapped_vectors)
        elif features == "hog":
            hf = compute_hog_features(gray_images)
            distance_matrix = scipy.spatial.distance.cdist(hf, hf)
        else:
            raise NotImplementedError(f"Features '{features}' are not implemented.")
        return distance_matrix

    def sort(self, gray_images: np.ndarray, features: str = "full", **kwargs) -> List[int]:
        """
        Sort images by posing the task as solving the Traveling Salesman Problem (TSP).
        We first compute pairwise distances between images. These are treated as a matrix which gives rise to an
        undirected weighted graph. By solving TSP on this graph, we obtain a Hamiltonian path (i.e. each vertex is
        theoretically visited only once), which is equal to the order of the sorted images. Note that an approximation
        of TSP is used and it is possible (likely) that an image may be encountered more than once.

        :param gray_images: input images.
        :param features: types of features to use when computing pairwise distances.
        :param kwargs: unused.
        :return: image indices in the final ordering.
        """
        self.verbose_print("Computing distance matrix")
        distance_matrix = self.distance_matrix(gray_images, features=features)
        self.verbose_print('Solving TSP')
        image_graph = nx.from_numpy_array(distance_matrix)
        order = nx.algorithms.approximation.traveling_salesman_problem(image_graph, cycle=False)
        return order


class HorizontalMassSorter(Sorter):
    def sort(self, gray_images: np.ndarray, **kwargs) -> List[int]:
        """
        Sort images using a horizontal center mass algorithm.
        For each image, we compute its center mass.
        Images are sorted according to the horizontal (x) component of the center mass.

        :param gray_images: input images.
        :param kwargs: unused.
        :return: image indices in the final ordering.
        """
        centers_x = []
        for image in gray_images:
            cy, cx = scipy.ndimage.center_of_mass(image)
            centers_x.append(cx)
        order = list(np.argsort(centers_x))
        return order


def resize_images(images: np.ndarray, target_size: Tuple[int, int] = (128, 128)) -> np.ndarray:
    """
    Resize images to a specified size.

    :param images: input images.
    :param target_size: target size as (height, width).
    :return: array of resized images.
    """
    return np.array([skimage.transform.resize(im, target_size) for im in images])


def rgb_to_gray(images):
    """
    Convert images from RGB to grayscale.

    :param images: input images with shape (num_images, height, width, 3).
    :return: grayscale images with shape (num_images, height, width).
    """
    return np.clip(
        (images[..., 0] * 0.299 + images[..., 1] * 0.587 + images[..., 2] * 0.114) / 255,
        0,
        1
    ).astype(np.float32)


def sort_images(images: np.ndarray,
                sort_type: str = "tsp",
                verbose: bool = False,
                working_size: Tuple[int, int] = None,
                **kwargs) -> List[int]:
    """
    Sort images with a specified sorting method.

    :param images: input images to sort.
    :param sort_type: method to sort images when constructing the animation. Should be one of the following:
        "tsp" (traveling salesman method on the image similarity graph), "pca" (image projection onto the real line
        using PCA), "hm" (order using center mass in the horizontal direction), "none" (no sorting; random order).
    :param verbose: if True, print additional information during sorting.
    :param working_size: optional (height, width) tuple that determines working image size. Images are resized to this
        size before being sorted.
    :param kwargs: keyword arguments for the sorting method.
    :return: image indices in the final ordering.
    """
    if sort_type == "tsp":
        sorter = TravelingSalesmanSorter(verbose=verbose)
    elif sort_type == "pca":
        sorter = PCASorter(verbose=verbose)
    elif sort_type == "hm":
        sorter = HorizontalMassSorter(verbose=verbose)
    elif sort_type == "none":
        sorter = DefaultSorter(verbose=verbose)
    else:
        raise NotImplementedError(f"Sort type '{sort_type}' not implemented")

    gray_images = rgb_to_gray(images)
    if working_size is not None:
        gray_images = resize_images(gray_images, working_size)

    order = sorter.sort(gray_images, **kwargs)
    return order



import numpy as np
from bootplot.image_features import compute_hog_features, compute_center_of_mass, compute_center_of_mass_per_component


def make_gray_images():
    np.random.seed(0)
    return np.random.uniform(low=0, high=1, size=(25, 100, 100))


def test_hog():
    images = make_gray_images()
    features = compute_hog_features(images)
    assert len(features.shape) == 2
    assert len(features) == len(images)

    small_images = images[:, :10, :10]
    features = compute_hog_features(small_images)
    assert len(features.shape) == 2
    assert len(features) == len(small_images)


def test_center_of_mass():
    images = make_gray_images()
    features = compute_center_of_mass(images)
    assert len(features.shape) == 2
    assert len(features) == len(images)

    small_images = images[:, :10, :10]
    features = compute_center_of_mass(small_images)
    assert len(features.shape) == 2
    assert len(features) == len(small_images)


def test_center_of_mass_per_component():
    images = make_gray_images()
    features = compute_center_of_mass_per_component(images)
    assert len(features.shape) == 2
    assert len(features) == len(images)

    small_images = images[:, :10, :10]
    features = compute_center_of_mass_per_component(small_images)
    assert len(features.shape) == 2
    assert len(features) == len(small_images)
